/* cmd.c, Boone, 09/29/04
   Command functions for talking to an xcat board
   Copyright (C) 2004-2009, Dennis Boone, East Lansing, MI */

/* Modifications:
   09/29/04 Boone      Initial coding
   End Modifications */

/* This file is part of xcatctl.

   Xcatctl is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License,
   or (at your option) any later version.

   Xcatctl is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with xcatctl; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "xcatctl.h"

int sendcmd(int to, int from, unsigned char *cmd, int cmdlen)
{
	unsigned char cfrom;
	unsigned char cto;
	int rc;
	int i;

	if (debug)
		fprintf(stderr, "sendcmd: to %02x from %02x\n", to, from);
	if ((from > 0xfc) || (to > 0xfc))
	{
		errno = EINVAL;
		return(EINVAL);
	}
	cfrom = from & 0xff;
	cto = to & 0xff;

	if (debug)
		fputs("sendcmd: sending header\n", stderr);
	rc = rawcharout(0xfe); if (rc) return(rc);
	rc = rawcharout(0xfe); if (rc) return(rc);
	rc = rawcharout(cto); if (rc) return(rc);
	rc = rawcharout(cfrom); if (rc) return(rc);
	if (debug)
		fputs("sendcmd: done sending header\n", stderr);

	for (i = 0; i < cmdlen; i++)
	{
		if (debug)
			fprintf(stderr, "sendcmd: sending cmd byte %d: %02x\n",
				i+1, cmd[i]);
		rc = charout(cmd[i]); if (rc) return(rc);
		if (debug)
		{
			fprintf(stderr, "sendcmd: charout returned %d\n", rc);
			if (rc != 0)
				perror("sendcmd");
		}
		if (rc) return(rc);
	}

	if (debug)
		fputs("sendcmd: sending trailer\n", stderr);
	rc = rawcharout(0xfd); if (rc) return(rc);
	if (debug)
		fputs("sendcmd: done sending trailer\n", stderr);
	if (rc) return(rc);

	return(0);
}

int rcvresp(int *to, int *from, unsigned char *resp, int bufsize, int *resplen)
{
	int rc;
	unsigned char c;
	int i;

	c = 0;
	while (c != 0xfe)
	{
		if (debug)
			fputs("rcvresp: eating non-header characters\n", stderr);
		rc = rawcharin(&c);
		if (debug)
		{
			fprintf(stderr, "rcvresp: rawcharin returned %d, eaten char %02x\n",
				rc, c);
			if (rc != 0)
				perror("rcvresp");
		}
		if (rc) return(rc);
	}
	if (debug)
		fputs("rcvresp: found beginning of header\n", stderr);
	rc = rawcharin(&c); if (rc) return(rc);
	if (c != 0xfe) return(EINVAL);
	rc = rawcharin(&c); if (rc) return(rc);
	*to = c;
	rc = rawcharin(&c); if (rc) return(rc);
	*from = c;
	if (debug)
		fprintf(stderr, "rcvresp: found full header, to=%02x from=%02x\n",
			*to, *from);

	if (debug)
		fputs("rcvresp: reading packet\n", stderr);
	i = 0;
	while (i < bufsize)
	{
		if (debug)
			fputs("rcvresp: reading character\n", stderr);
		rc = charin(&c); if (rc) return(rc);
		if (debug)
		{
			fprintf(stderr, "rcvresp: charin returned %d; c=%02x\n", rc, c);
			if (rc != 0)
				perror("rcvresp");
		}
		if (c == 0xfd)
		{
			if (debug)
				fputs("rcvresp: end of packet found\n", stderr);
			break;
		}
		resp[i] = c;
		i++;
	}
	if (debug)
		fprintf(stderr, "rcvresp: exited packet read loop; len=%d\n", i);

	*resplen = i;

	return(0);
}

int setvfofreq(int to, int from, long freq)
{
	unsigned char outbuf[6];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[6];
	int resplen;

	if (debug)
		fprintf(stderr, "setvfofreq: setting freq %7.3f\n", freq);
	outbuf[0] = 0x05;
	tobcdrev(freq, outbuf+1, 5);
	if (debug)
	{
		fputs("setvfofreq: bcd ", stderr);
		for (i = 0; i < 6; i++)
		{
			fprintf(stderr, "%02x ", outbuf[i]);
		}
		fputs("\n", stderr);
	}
	if (debug)
		fputs("setvfofreq: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 6);
	if (debug)
	{
		fprintf(stderr, "setvfofreq: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setvfofreq");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setvfofreq: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 6, &resplen);
		if (debug)
		{
			fprintf(stderr, "setvfofreq: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setvfofreq");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setvfofreq: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int setmode(int to, int from, long mode)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[2];
	int resplen;

	if (debug)
		fprintf(stderr, "setmode: selecting mode %d\n", mode);
	outbuf[0] = 0x08;
	tobcd(mode, outbuf+1, 1);
	if (debug)
	{
		fputs("setmode: bcd ", stderr);
		for (i = 0; i < 2; i++)
		{
			fprintf(stderr, "%02x ", outbuf[i]);
		}
		fputs("\n", stderr);
	}
	if (debug)
		fputs("setmode: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "setmode: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setmode");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setmode: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 2, &resplen);
		if (debug)
		{
			fprintf(stderr, "setmode: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setmode");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setmode: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int storevfotomode(int to, int from)
{
	unsigned char outbuf[2];
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[2];
	int resplen;

	if (debug)
		fputs("storevfotomode: entering\n", stderr);
	outbuf[0] = 0x09;
	if (debug)
		fputs("storevfotomode: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 1);
	if (debug)
	{
		fprintf(stderr, "storevfotomode: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("storevfotomode");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("storevfotomode: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 2, &resplen);
		if (debug)
		{
			fprintf(stderr, "storevfotomode: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("storevfotomode");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "storevfotomode: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int storemodetovfo(int to, int from)
{
	unsigned char outbuf[2];
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[2];
	int resplen;

	if (debug)
		fputs("storemodetovfo: entering\n", stderr);
	outbuf[0] = 0x0a;
	if (debug)
		fputs("storevfotomode: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 1);
	if (debug)
	{
		fprintf(stderr, "storemodetovfo: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("storemodetovfo");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("storemodetovfo: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 2, &resplen);
		if (debug)
		{
			fprintf(stderr, "storemodetovfo: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("storemodetovfo");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "storemodetovfo: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int setduplex(int to, int from, long duplex)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[2];
	int resplen;

	if (debug)
		fprintf(stderr, "setduplex: setting duplex to %02x\n", duplex);
	outbuf[0] = 0x0f;
	outbuf[1] = duplex;
	if (debug)
	{
		fputs("setduplex: bcd ", stderr);
		for (i = 0; i < 2; i++)
		{
			fprintf(stderr, "%02x ", outbuf[i]);
		}
		fputs("\n", stderr);
	}
	if (debug)
		fputs("setduplex: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "setduplex: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setduplex");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setduplex: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 2, &resplen);
		if (debug)
		{
			fprintf(stderr, "setduplex: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setduplex");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setduplex: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int setpl(int to, int from, long pl, int txrx)
{
	unsigned char outbuf[4];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[4];
	int resplen;

	if (debug)
		fprintf(stderr, "setpl: setting %s pl to %5.1f\n", 
			txrx ? "rx" : "tx", (float)pl / 10.0);
	outbuf[0] = 0x1b;
	outbuf[1] = txrx;
	tobcd(pl, outbuf+2, 2);
	if (debug)
	{
		fputs("setpl: bcd ", stderr);
		for (i = 0; i < 4; i++)
		{
			fprintf(stderr, "%02x ", outbuf[i]);
		}
		fputs("\n", stderr);
	}
	if (debug)
		fputs("setpl: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 4);
	if (debug)
	{
		fprintf(stderr, "setpl: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setpl");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setpl: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 4, &resplen);
		if (debug)
		{
			fprintf(stderr, "setpl: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setpl");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setpl: found our message; status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int getvfocodeplug(int to, int from, unsigned char *retbuf, int buflen)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[18];
	int resplen;

	if (debug)
		fputs("getvfocodeplug: entering\n", stderr);
	if (buflen < 16)
		return(ENOSPC);
	if (debug)
		fputs("getvfocodeplug: adequate buffer space is available\n", stderr);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x00;
	if (debug)
		fputs("getvfocodeplug: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "getvfocodeplug: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("getvfocodeplug");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("getvfocodeplug: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 18, &resplen);
		if (debug)
		{
			fprintf(stderr, "getvfocodeplug: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("getvfocodeplug");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "getvfocodeplug: found our message; "
			"command %02x subcommand %02x\n",
			inbuf[0], inbuf[1]);
	if ((inbuf[0] == 0xaa) && (inbuf[1] == 0x80))
	{
		if (debug)
		{
			fputs("getvfocodeplug: plugdata ", stderr);
			for (i = 0; i < 16; i++)
				fprintf(stderr, "%02x ", inbuf[i+2]);
			fputs("\n", stderr);
		}
		memcpy(retbuf, inbuf+2, 16);
		return(0);
	}
	errno = EINVAL;
	return(EINVAL);
}

int setvfocodeplug(int to, int from, char *plug)
{
	unsigned char outbuf[18];
	unsigned int hexbuf[16];
	int i;
	int rc;
	int rto = 0;
	int rfrom = 0;
	unsigned char inbuf[18];
	int resplen;

	if (debug)
		fprintf(stderr, "setvfocodeplug: plug to send [%s]\n", plug);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x01;
	rc = sscanf(plug,
		"%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
		&(hexbuf[0]), &(hexbuf[1]), &(hexbuf[2]), &(hexbuf[3]),
		&(hexbuf[4]), &(hexbuf[5]), &(hexbuf[6]), &(hexbuf[7]),
		&(hexbuf[8]), &(hexbuf[9]), &(hexbuf[10]), &(hexbuf[11]),
		&(hexbuf[12]), &(hexbuf[13]), &(hexbuf[14]), &(hexbuf[15]));
	if (rc != 16)
	{
		if (debug)
			fprintf(stderr, "setvfocodeplug: only got %d of 16 plug bytes\n",
				rc);
		errno = EINVAL;
		return(EINVAL);
	}
	for (i = 2; i < 18; i++)
		outbuf[i] = hexbuf[i-2];
	if (debug)
	{
		fputs("setvfocodeplug: parsed plug ", stderr);
		for (i = 0; i < 16; i++)
			fprintf(stderr, "%02x ", outbuf[i+2]);
		fputs("\n", stderr);
	}
	if (debug)
		fputs("setvfocodeplug: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 18);
	if (debug)
	{
		fprintf(stderr, "setvfocodeplug: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setvfocodeplug");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setvfocodeplug: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 18, &resplen);
		if (debug)
		{
			fprintf(stderr, "setvfocodeplug: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setvfocodeplug");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setvfocodeplug: found our message; "
			"status %02x\n", inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int getversion(int to, int from, unsigned char *retbuf, int buflen)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[256];
	int resplen;

	if (debug)
		fputs("getversion: entering\n", stderr);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x02;
	for (i = 0; i < 256; i++)
		inbuf[i] = 0;
	if (debug)
		fputs("getversion: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "getversion: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("getversion");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("getversion: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 255, &resplen);
		if (debug)
		{
			fprintf(stderr, "getversion: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("getversion");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "getversion: found our message; "
			"command %02x subcommand %02x\n",
			inbuf[0], inbuf[1]);
	if ((inbuf[0] == 0xaa) && (inbuf[1] == 0x82))
	{
		if (debug)
		{
			fputs("getversion: data ", stderr);
			for (i = 0; inbuf[i+2]; i++)
				fprintf(stderr, "%02x ", inbuf[i+2]);
			fputs("\n", stderr);
		}
		strncpy(retbuf, inbuf+2, buflen);
		return(0);
	}
	errno = EINVAL;
	return(EINVAL);
}

int getconfig(int to, int from, unsigned char *retbuf, int buflen)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[4];
	int resplen;

	if (debug)
		fputs("getconfig: entering\n", stderr);
	if (buflen < 2)
	{
		errno = ENOSPC;
		return(ENOSPC);
	}
	if (debug)
		fputs("getconfig: adequate buffer space is available\n", stderr);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x03;
	if (debug)
		fputs("getconfig: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "getconfig: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("getconfig");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("getconfig: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 4, &resplen);
		if (debug)
		{
			fprintf(stderr, "getconfig: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("getconfig");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "getconfig: found our message; "
			"command %02x subcommand %02x\n",
			inbuf[0], inbuf[1]);
	if ((inbuf[0] == 0xaa) && (inbuf[1] == 0x83))
	{
		if (debug)
		{
			fputs("getconfig: data ", stderr);
			for (i = 0; i < 2; i++)
				fprintf(stderr, "%02x ", inbuf[i+2]);
			fputs("\n", stderr);
		}
		memcpy(retbuf, inbuf+2, 2);
		return(0);
	}
	errno = EINVAL;
	return(EINVAL);
}

int setconfig(int to, int from, unsigned char data0, unsigned char data1)
{
	unsigned char outbuf[4];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[4];
	int resplen;

	if (debug)
		fprintf(stderr, "setconfig: config data0 %02x data1 %02x\n",
			data0, data1);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x04;
	outbuf[2] = data0;
	outbuf[3] = data1;
	if (debug)
		fputs("setconfig: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 4);
	if (debug)
	{
		fprintf(stderr, "setconfig: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setconfig");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setconfig: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 4, &resplen);
		if (debug)
		{
			fprintf(stderr, "setconfig: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setconfig");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setconfig: found our message; "
			"status %02x\n", inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int gettxoffset(int to, int from, long *offset)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[6];
	int resplen;
	long x;

	if (debug)
		fputs("gettxoffset: entering\n", stderr);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x05;
	if (debug)
		fputs("gettxoffset: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "gettxoffset: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("gettxoffset");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("gettxoffset: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 6, &resplen);
		if (debug)
		{
			fprintf(stderr, "gettxoffset: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("gettxoffset");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "gettxoffset: found our message; "
			"command %02x subcommand %02x\n",
			inbuf[0], inbuf[1]);
	if ((inbuf[0] == 0xaa) && (inbuf[1] == 0x85))
	{
		if (debug)
		{
			fputs("gettxoffset: data ", stderr);
			for (i = 0; i < 4; i++)
				fprintf(stderr, "%02x ", inbuf[i+2]);
			fputs("\n", stderr);
		}
		*offset = 0;
		for (i = 0; i < 4; i++)
		{
			*offset |= (inbuf[i+2] << (8*i));
		}
		return(0);
	}
	errno = EINVAL;
	return(EINVAL);
}

int settxoffset(int to, int from, long offset)
{
	unsigned char outbuf[6];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[6];
	int resplen;

	if (debug)
		fprintf(stderr, "settxoffset: offset %d\n", offset);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x06;
	for (i = 0; i < 4; i++)
		outbuf[i+2] = (offset & (0xff << (8*i))) >> (8*i);
	if (debug)
		fputs("settxoffset: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 6);
	if (debug)
	{
		fprintf(stderr, "settxoffset: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("settxoffset");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("settxoffset: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 6, &resplen);
		if (debug)
		{
			fprintf(stderr, "settxoffset: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("settxoffset");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "settxoffset: found our message; "
			"status %02x\n",
			inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}

int getvcosplitfreq(int to, int from, long *txsplit, long *rxsplit)
{
	unsigned char outbuf[2];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[10];
	int resplen;
	long x;
	long accum;

	if (debug)
		fputs("getvcosplit: entering\n", stderr);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x07;
	if (debug)
		fputs("getvcosplit: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 2);
	if (debug)
	{
		fprintf(stderr, "getvcosplit: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("getvcosplit");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("getvcosplit: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 10, &resplen);
		if (debug)
		{
			fprintf(stderr, "getvcosplit: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("getvcosplit");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "getvcosplit: found our message; "
			"command %02x subcommand %02x\n",
			inbuf[0], inbuf[1]);
	if ((inbuf[0] == 0xaa) && (inbuf[1] == 0x87))
	{
		if (debug)
		{
			fputs("getvcosplit: data ", stderr);
			for (i = 0; i < 8; i++)
				fprintf(stderr, "%02x ", inbuf[i+2]);
			fputs("\n", stderr);
		}
		*rxsplit = 0;
		*txsplit = 0;
		for (i = 0; i < 4; i++)
		{
			*rxsplit |= (inbuf[i+2] << (8*i));
			*txsplit |= (inbuf[i+6] << (8*i));
		}
		return(0);
	}
	errno = EINVAL;
	return(EINVAL);
}

int setvcosplit(int to, int from, long txvcosplit, long rxvcosplit)
{
	unsigned char outbuf[10];
	int i;
	int rto = 0;
	int rfrom = 0;
	int rc;
	unsigned char inbuf[10];
	int resplen;

	if (debug)
		fprintf(stderr, "setvcosplit: txsplit %d rxsplit %d\n",
			txvcosplit, rxvcosplit);
	outbuf[0] = 0xaa;
	outbuf[1] = 0x08;
	for (i = 0; i < 4; i++)
	{
		outbuf[i+2] = (rxvcosplit & (0xff << (8*i))) >> (8*i);
		outbuf[i+6] = (txvcosplit & (0xff << (8*i))) >> (8*i);
	}
	if (debug)
		fputs("setvcosplit: calling sendcmd\n", stderr);
	rc = sendcmd(to, from, outbuf, 10);
	if (debug)
	{
		fprintf(stderr, "setvcosplit: sendcmd returned %d\n", rc);
		if (rc != 0)
			perror("setvcosplit");
	}
	if (rc) return(rc);
	while ((rto != from) && (rfrom != to))
	{
		if (debug)
			fputs("setvcosplit: eating messages to/from other ids\n", stderr);
		rc = rcvresp(&rto, &rfrom, inbuf, 2, &resplen);
		if (debug)
		{
			fprintf(stderr, "setvcosplit: rcvresp returned to %02x from "
				"%02x rc %d\n",
				rto, rfrom, rc);
			if (rc != 0)
				perror("setvcosplit");
		}
		if (rc) return(rc);
	}
	if (debug)
		fprintf(stderr, "setvcosplit: found our message; "
			"status %02x\n", inbuf[0]);
	if (inbuf[0] == 0xfb) return(0);
	errno = EINVAL;
	return(EINVAL);
}
