/* cosmon.c, Boone, 09/29/04
   Listen for carrier status messages from an xcat board
   Copyright (C) 2004-2009, Dennis Boone, East Lansing, MI */

/* Modifications:
   09/29/04 Boone      Initial coding
   End Modifications */

/* This file is part of xcatctl.

   Xcatctl is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License,
   or (at your option) any later version.

   Xcatctl is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with xcatctl; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA */

#define _GNU_SOURCE
#define INSTANTIATE

#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/select.h>
#include <signal.h>
#include "cosmon.h"
#include "cosopts.h"
#include "brate.h"
#include "cosfsa.h"

int cosfsar(int state, int input, int toaddr, int fromaddr)
{
	int i;

	for (i = 0; cosfsa[i].state != -1; i++)
	{
		if (debug)
			fprintf(stderr, "cosfsa: state=%d input=%02x [%d]={%d,%02x,%d}\n",
				state, input, i, cosfsa[i].state,
				cosfsa[i].input, cosfsa[i].dest);
		if (cosfsa[i].state == state)
		{
			if (cosfsa[i].input == (unsigned int)input)
			{
				if (debug)
					fputs("cosfsa: input matched\n", stderr);
				return(cosfsa[i].dest);
			}
			else
				switch (cosfsa[i].input)
				{
					case -4:
						if ((input >= 0x00) && (input <= 0x1f))
						{
							if (debug)
								fputs("cosfsa: valid mode number matched\n",
									stderr);
							return(cosfsa[i].dest);
						}
						break;
					case -3:
						if (input == toaddr)
						{
							if (debug)
								fputs("cosfsa: toaddr matched\n", stderr);
							return(cosfsa[i].dest);
						}
						break;
					case -2:
						if (input == fromaddr)
						{
							if (debug)
								fputs("cosfsa: fromaddr matched\n", stderr);
							return(cosfsa[i].dest);
						}
						break;
					case -1:
						if (debug)
							fputs("cosfsa: catchall matched\n", stderr);
						return(cosfsa[i].dest);
						break;
				}
		}
	}
}

#ifdef __CYGWIN__
void siglost(char *dncmd, char *serialport, unsigned char toaddr)
{
	int pid;
	char stoaddr[256];
	int rc;

	if (debug)
		fprintf(stderr, "siglost: invoking dncmd %s\n", dncmd);

	pid = fork();
	if (pid)
		return;

	snprintf(stoaddr, 255, "0x%02x", toaddr);
	rc = execl(dncmd, dncmd, serialport, stoaddr, NULL);
	if (debug)
		fprintf(stderr, "siglost: rc %d, errno %d from execl\n", rc, errno);
	if (rc == -1)
		perror("siglost");
	exit(0);
}

void sigfound(char *upcmd, char *serialport, unsigned char toaddr)
{
	int pid;
	char stoaddr[256];
	int rc;

	if (debug)
		fprintf(stderr, "sigfound: invoking upcmd %s\n", upcmd);

	pid = fork();
	if (pid)
		return;

	snprintf(stoaddr, 255, "0x%02x", toaddr);
	rc = execl(upcmd, upcmd, serialport, stoaddr, NULL);
	if (debug)
		fprintf(stderr, "sigfound: rc %d, errno %d from execl\n", rc, errno);
	if (rc == -1)
		perror("sigfound");
	exit(0);
}
#else
void siglost(char *dncmd, char *serialport, unsigned char toaddr)
{
	int pid;
	char cmdbuf[1024];
	int rc;

	if (debug)
		fprintf(stderr, "siglost: invoking dncmd %s\n", dncmd);

	pid = fork();
	if (pid)
		return;

	snprintf(cmdbuf, 1023, "%s %s 0x%02x", dncmd, serialport, toaddr);
	rc = system(cmdbuf);
	if (debug)
		fprintf(stderr, "siglost: rc %d, errno %d from system\n", rc, errno);
	if (rc == -1)
		perror("siglost");
	exit(0);
}

void sigfound(char *upcmd, char *serialport, unsigned char toaddr)
{
	int pid;
	char cmdbuf[1024];
	int rc;

	if (debug)
		fprintf(stderr, "sigfound: invoking upcmd %s\n", upcmd);

	pid = fork();
	if (pid)
		return;

	snprintf(cmdbuf, 1023, "%s %s 0x%02x", upcmd, serialport, toaddr);
	rc = system(cmdbuf);
	if (debug)
		fprintf(stderr, "sigfound: rc %d, errno %d from system\n", rc, errno);
	if (rc == -1)
		perror("sigfound");
	exit(0);
}
#endif

int main(int argc, char *argv[])
{
	char serialport[256];
	int rc;
	int longindex;
	int fail = 0;
	char upcmd[1024] = "";
	int doup = 0;
	char dncmd[1024] = "";
	int dodn = 0;
	char ptydev[256] = "";
	int dopty = 0;
#ifdef __CYGWIN__
	char ttydev[256] = "";
#endif
	unsigned char toaddr = 0x20;
	unsigned char fromaddr = 0xe0;
	int baudrate = B19200;
	int brfound;
	int i;
	int serfh = -2;
	struct termios tp;
	int ptyfh;
	FILE *ptsfh;
	int maxfh;
	fd_set serset;
	struct timeval tv;
	int nd;
	int nc;
	unsigned char c;
	int nw;
	struct sigaction sa;
	int state;
	int haveclient;

/* Initialize */

	verbose = 0;
	debug = 0;
	strcpy(serialport, "/dev/ttyS0");

/* Command line */

	while ((rc = getopt_long(argc, argv, "", longopts, &longindex)) != -1)
	{
		if (rc == '?')
		{
			fail = 1;
			continue;
		}
		else
		{
			if (debug)
				fprintf(stderr, "cosmon: --%s\n", longopts[longindex].name);
			switch (rc)
			{
				case UPCMD:
					strncpy(upcmd, optarg, 1023);
					doup = 1;
					break;
				case DNCMD:
					strncpy(dncmd, optarg, 1023);
					dodn = 1;
					break;
#ifndef __CYGWIN__
				case PTYDEV:
					strncpy(ptydev, optarg, 255);
					dopty = 1;
					break;
#endif
#ifdef __CYGWIN__
				case WINDOWS:
					strncpy(ttydev, optarg, 255);
					dopty = 1;
					break;
#endif
				case TOADDR:
					toaddr = strtod(optarg, NULL);
					break;
				case FROMADDR:
					fromaddr = strtod(optarg, NULL);
					break;
				case DEBUG:
					debug = 1;
					break;
				case SERIALPORT:
					strncpy(serialport, optarg, 255);
					break;
				case BAUDRATE:
					baudrate = atoi(optarg);
					brfound = 0;
					for (i = 0; brmap[i].baudrate != -1; i++)
					{
						if (brmap[i].baudrate == baudrate)
						{
							baudrate = brmap[i].bflag;
							brfound = 1;
							break;
						}
					}
					if (! brfound)
					{
						fprintf(stderr, "cosmon: invalid baud rate %s\n",
							optarg);
						fail = 1;
					}
					break;
				case HELP:
					help();
					exit(0);
					break;
				case SHOWVERSION:
					printf("cosmon version %s\n", VERSION);
					exit(0);
					break;
				default:
					fprintf(stderr, "cosmon: logic error in parser, index %d\n",
						rc);
			}
		}
	}

	if ((! doup) && (! dodn))
	{
		fputs("cosmon: you must specify at least one of --upcommand or "
			"--downcommand\n",
			stderr);
		fail = 1;
	}

	if (fail)
		exit(EINVAL);

/* Set up the serial port */

	if (debug)
		fprintf(stderr, "cosmon: opening device %s\n", serialport);
#ifdef __CYGWIN__
	serfh = open(serialport, O_RDWR|O_NOCTTY|O_BINARY, 0);
#else
	serfh = open(serialport, O_RDWR|O_NOCTTY, 0);
#endif
	if (debug)
		fprintf(stderr, "cosmon: open of serial port returned fh %d\n", serfh);
	if (serfh == -1)
	{
		perror("cosmon");
		exit(errno);
	}

	if (debug)
		fputs("cosmon: getting current serial device attributes\n", stderr);
	rc = tcgetattr(serfh, &tp);
	if (debug)
		fprintf(stderr, "cosmon: tcgetattr returned %d\n", rc);
	if (rc)
		exit(errno);
	tp.c_iflag = (IGNBRK|IGNPAR);
	tp.c_oflag = 0;
	tp.c_cflag = (CS8|CREAD|HUPCL|CLOCAL);
	tp.c_lflag = 0;
	rc = cfsetispeed(&tp, baudrate);
	if (rc)
		exit(errno);
	rc = cfsetospeed(&tp, baudrate);
	if (rc)
		exit(errno);
/*	cfmakeraw(&tp); */
	if (debug)
		fputs("cosmon: setting serial device attributes\n", stderr);
	rc = tcsetattr(serfh, 0, &tp);
	if (debug)
	{
		fprintf(stderr, "cosmon: tcsetattr returned %d\n", rc);
		if (rc == -1)
			perror("cosmon: open serial");
	}
	if (rc)return(errno);

/* Set up the pass-through pty, if any */

	if (dopty)
	{
		if (debug)
			fputs("cosmon: opening pty\n", stderr);
#ifndef __CYGWIN__
		ptyfh = open(ptydev, O_RDWR|O_NOCTTY, 0);
#endif
#ifdef __CYGWIN__
		ptyfh = open("/dev/ptmx", O_RDWR|O_BINARY, 0);
		if (ptyfh == -1)
		{
			perror("cosmon: open pty");
			exit(errno);
		}
		rc = grantpt(ptyfh);
		if (rc < 0)
		{
			perror("cosmon: grantpt");
			exit(errno);
		}
		rc = unlockpt(ptyfh);
		if (rc < 0)
		{
			perror("cosmon: unlockpt");
			exit(errno);
		}
		if ((ptsfh = fopen(ttydev, "w")) == NULL)
		{
			perror("cosmon: open ttydev");
			exit(errno);
		}
		fprintf(ptsfh, "%s", ptsname(ptyfh));
		fclose(ptsfh);
#endif
		if (debug)
			fprintf(stderr, "cosmon: ptyfh %d from open of ptydev\n", ptyfh);
		if (ptyfh == -1)
		{
			perror("cosmon: open ptydev");
			exit(errno);
		}
		if (debug)
			fputs("cosmon: getting current pty device attributes\n", stderr);
		rc = tcgetattr(ptyfh, &tp);
		if (debug)
			fprintf(stderr, "cosmon: tcgetattr returned %d\n", rc);
		if (rc)
			exit(errno);
		tp.c_iflag = (IGNBRK|IGNPAR);
		tp.c_oflag = 0;
		tp.c_cflag = (CS8|CREAD|HUPCL|CLOCAL);
		tp.c_lflag = 0;
		rc = cfsetispeed(&tp, baudrate);
		if (rc)
			exit(errno);
		rc = cfsetospeed(&tp, baudrate);
		if (rc)
			exit(errno);
/*		cfmakeraw(&tp); */
		if (debug)
			fputs("cosmon: setting pty device attributes\n", stderr);
		rc = tcsetattr(ptyfh, 0, &tp);
		if (debug)
		{
			fprintf(stderr, "cosmon: tcsetattr returned %d\n", rc);
			if (rc == -1)
				perror("cosmon: open serial");
		}
		if (rc)return(errno);
	}

/* Daemonize */

	if (! debug)
	{
/*		fclose(stdin);
		fclose(stdout);
		fclose(stderr); */

		rc = fork();
		if (rc == -1)
		{
			perror("cosmon");
			exit(errno);
		}
		if (rc)
			exit(0);

/*		setsid(); */
	}

/* Ignore dying children */

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	rc = sigaction(SIGCHLD, &sa, NULL);

/* Main loop */

	state = 0;
	haveclient = 0;
	while (1)
	{
		if (dopty)
			maxfh = (ptyfh > serfh) ? ptyfh : serfh;
		else
			maxfh = serfh;
		FD_ZERO(&serset);
		FD_SET(serfh, &serset);
		if (dopty)
			FD_SET(ptyfh, &serset);
		tv.tv_sec = 30;
		tv.tv_usec = 0;
		if (debug)
			fputs("cosmon: calling select\n", stderr);
		nd = select(maxfh+1, &serset, NULL, NULL, &tv);
		if (debug)
		{
			fprintf(stderr, "cosmon: select returned %d\n", nd);
			if (nd == -1)
				perror("rawcharin");
		}
		if (nd == 0)
			continue;

		if (FD_ISSET(serfh, &serset))
		{
			if (debug)
				fputs("cosmon: reading a character from serial\n", stderr);
			nc = read(serfh, &c, 1);
			if (debug)
			{
				fprintf(stderr, "cosmon: read returned %d, c=%02x\n", nc, c);
				if (nc == -1)
					perror("cosmon");
			}
			if (nc == -1)
				return(errno);
			if (nc == 0)
			{
				errno = EPIPE;
				return(EPIPE);
			}

			if (dopty && haveclient)
			{
				nw = 0;
				if (debug)
					fprintf(stderr, "cosmon: writing char %02x to pty\n", c);
				while (! nw)
					nw = write(ptyfh, &c, 1);
			}

			state = cosfsar(state, (int)c, toaddr, fromaddr);
			if (debug)
				fprintf(stderr, "cosmon: new state is %d\n", state);
			if ((state == 10) && dodn)
				siglost(dncmd, serialport, toaddr);
			if ((state == 11) && doup)
				sigfound(upcmd, serialport, toaddr);
		}

		if (FD_ISSET(ptyfh, &serset))
		{
			haveclient = 1;
			if (debug)
				fputs("cosmon: reading a character from pty\n", stderr);
			nc = read(ptyfh, &c, 1);
			if (debug)
			{
				fprintf(stderr, "cosmon: read returned %d, c=%02x\n", nc, c);
				if (nc == -1)
					perror("cosmon");
			}
			if (nc < 1)
			{
				haveclient = 0;
				close(ptyfh);
				if (debug)
					fputs("cosmon: reopening pty\n", stderr);
#ifndef __CYGWIN__
				ptyfh = open(ptydev, O_RDWR|O_NOCTTY, 0);
#endif
#ifdef __CYGWIN__
				ptyfh = open("/dev/ptmx", O_RDWR|O_BINARY, 0);
				if ((ptsfh = fopen(ttydev, "w")) == NULL)
				{
					perror("cosmon: open ptydev");
					exit(errno);
				}
				rc = grantpt(ptyfh);
				if (rc < 0)
				{
					perror("cosmon: grantpt");
					exit(errno);
				}
				rc = unlockpt(ptyfh);
				if (rc < 0)
				{
					perror("cosmon: unlockpt");
					exit(errno);
				}
				fprintf(ptsfh, "%s", ptsname(ptyfh));
				fclose(ptsfh);
#endif
				if (debug)
					fprintf(stderr,
						"cosmon: ptyfh %d from open of ptydev\n", ptyfh);
				if (ptyfh == -1)
				{
					perror("cosmon: open ptydev");
					exit(errno);
				}
				if (debug)
					fputs("cosmon: getting current pty device attributes\n", stderr);
				rc = tcgetattr(ptyfh, &tp);
				if (debug)
					fprintf(stderr, "cosmon: tcgetattr returned %d\n", rc);
				if (rc)
					exit(errno);
				tp.c_iflag = (IGNBRK|IGNPAR);
				tp.c_oflag = 0;
				tp.c_cflag = (CS8|CREAD|HUPCL|CLOCAL);
				tp.c_lflag = 0;
				rc = cfsetispeed(&tp, baudrate);
				if (rc)
					exit(errno);
				rc = cfsetospeed(&tp, baudrate);
				if (rc)
					exit(errno);
		/*		cfmakeraw(&tp); */
				if (debug)
					fputs("cosmon: setting pty device attributes\n", stderr);
				rc = tcsetattr(ptyfh, 0, &tp);
				if (debug)
				{
					fprintf(stderr, "cosmon: tcsetattr returned %d\n", rc);
					if (rc == -1)
						perror("cosmon: open serial");
				}
				if (rc)return(errno);
						continue;
			}

			nw = 0;
			if (debug)
				fprintf(stderr, "cosmon: writing char %02x to serial\n", c);
			while (! nw)
				nw = write(serfh, &c, 1);
		}
	}

/* Done */

	exit(0);
}
