/* io.c, Boone, 09/29/04
   Input/output functions for talking to an xcat board
   Copyright (C) 2004-2009, Dennis Boone, East Lansing, MI */

/* Modifications:
   09/29/04 Boone      Initial coding
   End Modifications */

/* This file is part of xcatctl.

   Xcatctl is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License,
   or (at your option) any later version.

   Xcatctl is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with xcatctl; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/select.h>
#include "xcatctl.h"

int serfh = -2;

int seropen(char *devname, int baudrate)
{
	int rc;
	struct termios tp;

	if (debug)
		fprintf(stderr, "seropen: opening device %s\n", devname);
	serfh = open(devname, O_RDWR|O_NOCTTY, 0);
	if (debug)
	{
		fprintf(stderr, "seropen: open returned fh %d\n", serfh);
		if (serfh == -1)
			perror("seropen");
	}
	if (serfh == -1)
		return(errno);

	if (debug)
		fputs("seropen: getting current serial device attributes\n", stderr);
	rc = tcgetattr(serfh, &tp);
	if (debug)
		fprintf(stderr, "tcgetattr returned %d\n", rc);
	if (rc) return(errno);
	tp.c_iflag = (IGNBRK|IGNPAR);
	tp.c_oflag = 0;
	tp.c_cflag = (CS8|CREAD|HUPCL|CLOCAL);
	tp.c_lflag = 0;
	rc = cfsetispeed(&tp, baudrate);
	if (rc) return(errno);
	rc = cfsetospeed(&tp, baudrate);
	if (rc) return(errno);
/*	cfmakeraw(&tp); */
	if (debug)
		fputs("seropen: setting serial device attributes\n", stderr);
	rc = tcsetattr(serfh, 0, &tp);
	if (debug)
	{
		fprintf(stderr, "tcsetattr returned %d\n", rc);
		if (rc == -1)
			perror("seropen");
	}
	if (rc)return(errno);

	return(0);
}

int serclose()
{
	if (serfh > -1)
	{
		if (debug)
			fputs("serclose: closing serial device\n", stderr);
		close(serfh);
	}
}

int rawcharout(unsigned char c)
{
	ssize_t nc;

	if (debug)
		fprintf(stderr, "rawcharout: writing %02x\n", c);
	nc = write(serfh, &c, 1);
	if (debug)
	{
		fprintf(stderr, "rawcharout: write returned %d\n", nc);
		if (nc == -1)
			perror("rawcharout");
	}

	if (nc != 1)
		return(errno);
	else
		return(0);
}

int rawcharin(unsigned char *c)
{
	fd_set serset;
	struct timeval tv;
	int nd;
	ssize_t nc;

	FD_ZERO(&serset);
	FD_SET(serfh, &serset);
	tv.tv_sec = 5;
	tv.tv_usec = 0;
	if (debug)
		fputs("rawcharin: calling select\n", stderr);
	nd = select(serfh+1, &serset, NULL, NULL, &tv);
	if (debug)
	{
		fprintf(stderr, "rawcharin: select returned %d\n", nd);
		if (nd == -1)
			perror("rawcharin");
	}
	if (nd == 0)
	{
		errno = ETIMEDOUT;
		return(ETIMEDOUT);
	}

	if (debug)
		fputs("rawcharin: reading a character\n", stderr);
	nc = read(serfh, c, 1);
	if (debug)
	{
		fprintf(stderr, "rawcharin: read returned %d\n", nc);
		if (nc == -1)
			perror("rawcharin");
	}

	if (nc == -1)
		return(errno);
	if (nc == 0)
	{
		errno = EPIPE;
		return(EPIPE);
	}
	return(0);
}

int charout(unsigned char c)
{
	int rc;

	if ((c == 0xfd) || (c == 0xfe) || (c == 0xff))
	{
		if (debug)
			fputs("charout: escaping; rawcharout(0xff)\n", stderr);
		rc = rawcharout(0xff);
		if (debug)
		{
			fprintf(stderr, "charout: rawcharout returned %d\n", rc);
			if (rc != 0)
				perror("charout");
		}
		if (rc) return(rc);
		if (debug)
			fprintf(stderr, "charout: escaping; rawcharout(%02x)\n", c&0x0f);
		rc = rawcharout(c & 0x0f);
		if (debug)
		{
			fprintf(stderr, "charout: rawcharout returned %d\n", rc);
			if (rc != 0)
				perror("charout");
		}
		if (rc) return(rc);
	}
	else
	{
		if (debug)
			fprintf(stderr, "charout: not escaping; rawcharout(%02x)\n", c);
		rc = rawcharout(c);
		if (debug)
		{
			fprintf(stderr, "charout: rawcharout returned %d\n", rc);
			if (rc != 0)
				perror("charout");
		}
		if (rc) return(rc);
	}
	return(0);
}

int charin(unsigned char *c)
{
	int rc;

	if (debug)
		fputs("charin: calling rawcharin\n", stderr);
	rc = rawcharin(c);
	if (debug)
	{
		fprintf(stderr, "charin: rawcharin returned %d, c=%02x\n", rc, *c);
		if (rc != 0)
			perror("charout");
	}
	if (rc) return(rc);
	if (*c == 0xff)
	{
		if (debug)
			fputs("charin: escaping; calling rawcharin\n", stderr);
		rc = rawcharin(c);
		if (debug)
		{
			fprintf(stderr, "charin: rawcharin returned %d, c=%02x\n", rc, *c);
			if (rc != 0)
				perror("charout");
		}
		if (rc) return(rc);
		*c |= 0xf0;
	}
	return(0);
}
