/* xcatctl.c, Boone, 09/29/04
   Control an xcat board
   Copyright (C) 2004-2009, Dennis Boone, East Lansing, MI */

/* Modifications:
   09/29/04 Boone      Initial coding
   End Modifications */

/* This file is part of xcatctl.

   Xcatctl is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published
   by the Free Software Foundation; either version 2 of the License,
   or (at your option) any later version.

   Xcatctl is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with xcatctl; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA */

#define _GNU_SOURCE
#define INSTANTIATE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <math.h>
#include <termios.h>
#include "xcatctl.h"
#include "opts.h"
#include "brate.h"

void decodeconfig(unsigned char data0, unsigned char data1)
{
	int i;

	switch(data0 & 0x03)
	{
		case 0:
			fputs("Radio band: 10 meters\n", stdout);
			break;
		case 1:
			fputs("Radio band: 6 meters\n", stdout);
			break;
		case 2:
			fputs("Radio band: 10 and 6 meters\n", stdout);
			break;
		case 3:
			fputs("Radio band: 70 centimeters (also high band)\n", stdout);
			break;
	}

	switch((data0 & 0x04) >> 2)
	{
		case 0:
			fputs("Do not send COS messages\n", stdout);
			break;
		case 1:
			fputs("Send COS messages\n", stdout);
			break;
	}

	switch((data0 & 0x30) >> 4)
	{
		case 0:
			fputs("Control system: None (CI-V is always enabled)\n", stdout);
			break;
		case 1:
			fputs("Control system: Doug Hall / Generic\n", stdout);
			break;
		case 2:
			fputs("Control system: Palomar / Cactus\n", stdout);
			break;
		default:
			fputs("Invalid control system setting\n", stdout);
	}

	for (i = 0; i < 8; i++)
	{
		switch((data1 & (1 << (7-i))) >> (7-i))
		{
			case 0:
				printf("IO%d - alternate function\n", i);
				break;
			case 1:
				printf("IO%d - user function\n", i);
				break;
		}
	}
}

int main(int argc, char *argv[])
{
	int rc;
	int longindex;
	int fail = 0;
	int actions = 0;
	Cmdlopts cmdlopts[NUMACTIONS];
	int i;
	int to = 0x20;
	int from = 0xe0;
	char serialport[255];
	int baudrate = B19200;
	int brfound = 1;
	unsigned char inbuf[256];
	unsigned char a, b;
	long offset;
	long txsplit, rxsplit;
	long confdata0, confdata1;
	int validpl;
	float validpls[] =
	{
		67.0, 69.3, 71.9, 74.4, 77.0, 79.7, 82.5, 85.4, 88.5, 91.5, 94.8,
		97.4, 100.0, 103.5, 107.2, 110.9, 114.8, 118.8, 123.0, 127.3,
		131.8, 136.5, 141.3, 146.2, 151.4, 156.7, 162.2, 167.9, 173.8,
		179.9, 186.2, 192.8, 203.5, 206.5, 210.7, 218.1, 225.7, 229.1,
		233.6, 241.8, 250.3, 254.1, 0.0
	};


/* Initialize */

	verbose = 0;
	debug = 0;
	strcpy(serialport, "/dev/ttyS0");
	confdata0 = 0x03;
	confdata1 = 0xff;

	for (i = 0; i < NUMACTIONS; i++)
	{
		cmdlopts[i].oflag = 0;
		cmdlopts[i].havearg = 0;
	}

/* Command line */

	while ((rc = getopt_long(argc, argv, "", longopts, &longindex)) != -1)
	{
		if (rc == '?')
		{
			fail = 1;
			continue;
		}
		else
		{
			if (debug)
				fprintf(stderr, "xcatctl: --%s\n", longopts[longindex].name);
			switch (rc)
			{
				case VERBOSE:
					verbose = 1;
					break;
				case SETVFOFREQ:
					actions++;
					cmdlopts[rc].farg = strtod(optarg, NULL);
					cmdlopts[rc].iarg =
						(long)floor((cmdlopts[rc].farg * 1000000.0L));
					cmdlopts[rc].oflag = 1;
					break;
				case SETMODE:
					actions++;
					cmdlopts[rc].iarg = (int)strtol(optarg, NULL, 0) - 1;
					if ((cmdlopts[rc].iarg < 0) || (cmdlopts[rc].iarg > 31))
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"between 1 and 32\n", longopts[SETMODE].name);
						fail = 1;
					}
					cmdlopts[rc].oflag = 1;
					break;
				case STOREVFOTOMODE:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case STOREMODETOVFO:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETDUPLEX:
					actions++;
					if (strncasecmp(optarg, "simplex", strlen(optarg)) == 0)
					{
						cmdlopts[rc].iarg = 0x10;
					}
					else if (strncasecmp(optarg, "negative",
						strlen(optarg)) == 0)
					{
						cmdlopts[rc].iarg = 0x11;
					}
					else if (strncasecmp(optarg, "positive",
						strlen(optarg)) == 0)
					{
						cmdlopts[rc].iarg = 0x12;
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'simplex', 'negative' or 'positive'\n",
							longopts[SETDUPLEX].name);
						fail = 1;
					}
					cmdlopts[rc].oflag = 1;
					break;
				case SETTXPL:
					actions++;
					cmdlopts[rc].farg = strtod(optarg, NULL);
					validpl = 0;
					for (i = 0; validpls[i] != 0.0; i++)
						if (validpls[i] == cmdlopts[rc].farg)
						{
							validpl = 1;
							break;
						}
					if (!validpl)
						fprintf(stderr, "xcatctl: %5.1f is not a valid pl; "
							"will try anyway\n", cmdlopts[rc].farg);
					cmdlopts[rc].iarg = round(cmdlopts[rc].farg * 10.0L);
					cmdlopts[rc].oflag = 1;
					break;
				case SETRXPL:
					actions++;
					cmdlopts[rc].farg = strtod(optarg, NULL);
					cmdlopts[rc].iarg = (long)(cmdlopts[rc].farg * 10.0);
					cmdlopts[rc].oflag = 1;
					break;
				case GETVFOCODEPLUG:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETVFOCODEPLUG:
					actions++;
					cmdlopts[rc].oflag = 1;
					strncpy(cmdlopts[rc].sarg, optarg, 255);
					break;
				case GETVERSION:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case GETCONFIG:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETCONFIG:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETCONFIGBITS:
					actions++;
					cmdlopts[rc].iarg = (int)strtol(optarg, NULL, 0);
					cmdlopts[rc].oflag = 1;
					break;
				case GETTXOFFSET:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETTXOFFSET:
					actions++;
					cmdlopts[rc].iarg = (int)strtol(optarg, NULL, 0);
					cmdlopts[rc].oflag = 1;
					break;
				case GETVCOSPLIT:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case SETVCOSPLIT:
					actions++;
					cmdlopts[rc].oflag = 1;
					break;
				case TXVCOSPLIT:
					cmdlopts[rc].farg = strtod(optarg, NULL);
					cmdlopts[rc].iarg =
						(long)floor((cmdlopts[rc].farg * 1000000.0L));
					cmdlopts[rc].oflag = 1;
					break;
				case RXVCOSPLIT:
					cmdlopts[rc].farg = strtod(optarg, NULL);
					cmdlopts[rc].iarg =
						(long)floor((cmdlopts[rc].farg * 1000000.0L));
					cmdlopts[rc].oflag = 1;
					break;
				case RADIOBAND:
					if (strncasecmp(optarg, "10", strlen(optarg)) == 0)
					{
						confdata0 &= 0xf8;
					}
					else if (strncasecmp(optarg, "10+6", strlen(optarg)) == 0)
					{
						confdata0 &= 0xf8;
						confdata0 |= 0x02;
					}
					else if (strncasecmp(optarg, "6", strlen(optarg)) == 0)
					{
						confdata0 &= 0xf8;
						confdata0 |= 0x01;
					}
					else if (strncasecmp(optarg, "440", strlen(optarg)) == 0)
					{
						confdata0 &= 0xf8;
						confdata0 |= 0x03;
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'10', '10+6', '6' or '440'\n",
							longopts[RADIOBAND].name);
						fail = 1;
					}
					break;
				case SENDCOS:
					if (strncasecmp(optarg, "yes", strlen(optarg)) == 0)
					{
						confdata0 |= 0x08;
					}
					else if (strncasecmp(optarg, "no", strlen(optarg)) == 0)
					{
						confdata0 &= 0xf7;
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'yes' or 'no'\n",
							longopts[SENDCOS].name);
						fail = 1;
					}
					break;
				case CONTROLSYS:
					if (strncasecmp(optarg, "none", strlen(optarg)) == 0)
					{
						confdata0 &= 0x0f;
					}
					else if ((strncasecmp(optarg, "doughall",
								strlen(optarg)) == 0) ||
							(strncasecmp(optarg, "generic",
								strlen(optarg)) == 0))
					{
						confdata0 &= 0x0f;
						confdata0 |= 0x10;
					}
					else if ((strncasecmp(optarg, "palomar",
								strlen(optarg)) == 0) ||
							(strncasecmp(optarg, "cactus",
								strlen(optarg)) == 0))
					{
						confdata0 &= 0x0f;
						confdata0 |= 0x20;
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'none', 'doughall', 'generic', 'palomar' or "
							"'cactus'\n",
							longopts[CONTROLSYS].name);
						fail = 1;
					}
					break;
				case IO0:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x01;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x01);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO1:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x02;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x02);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO2:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x04;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x04);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO3:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x08;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x08);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO4:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x10;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x10);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO5:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x20;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x20);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO6:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x40;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x40);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case IO7:
					if (strncasecmp(optarg, "user", strlen(optarg)) == 0)
					{
						confdata1 |= 0x80;
					}
					else if (strncasecmp(optarg, "alternate",
							strlen(optarg)) == 0)
					{
						confdata1 &= (~ 0x80);
					}
					else
					{
						fprintf(stderr, "xcatctl: --%s argument must be "
							"'user'or 'alternate'\n",
							longopts[IO0].name);
						fail = 1;
					}
					break;
				case TOADDR:
					to = atoi(optarg);
					break;
				case FROMADDR:
					from = atoi(optarg);
					break;
				case DEBUG:
					debug = 1;
					break;
				case SERIALPORT:
					strncpy(serialport, optarg, 255);
					break;
				case BAUDRATE:
					baudrate = atoi(optarg);
					brfound = 0;
					for (i = 0; brmap[i].baudrate != -1; i++)
					{
						if (brmap[i].baudrate == baudrate)
						{
							baudrate = brmap[i].bflag;
							brfound = 1;
							break;
						}
					}
					if (! brfound)
					{
						fprintf(stderr, "xcatctl: invalid baud rate %s\n",
							optarg);
						exit(EINVAL);
					}
					break;
				case HELP:
					help();
					exit(0);
					break;
				case SHOWVERSION:
					printf("xcatctl version %s\n", VERSION);
					exit(0);
					break;
				default:
					fprintf(stderr, "xcatctl: logic error in command line "
						"parser, index %d\n", rc);
					fail = 1;
			};
		}
	}

	if (argv[optind] != NULL)
	{
		fprintf(stderr, "xcatctl: unrecognized option index %d value %s\n",
			optind, argv[optind]);
		exit(EINVAL);
	}

	if (fail)
		exit(EINVAL);

/* Take specified actions */

	if (actions)
	{
		rc = seropen(serialport, baudrate);
		if (rc)
		{
			fputs("xcatctl: ", stderr);
			perror(serialport);
			exit(rc);
		}
	}

	for (i = 1; i < NUMACTIONS; i++)
	{
		if (cmdlopts[i].oflag)
		{
			switch (i)
			{
				case VERBOSE:
					break;
				case SETVFOFREQ:
					rc = setvfofreq(to, from, cmdlopts[i].iarg);
					if (rc) perror("xcatctl: setvfofreq");
					break;
				case SETMODE:
					rc = setmode(to, from, cmdlopts[i].iarg);
					if (rc) perror("xcatctl: setmode");
					break;
				case STOREVFOTOMODE:
					rc = storevfotomode(to, from);
					if (rc) perror("xcatctl: storevfotomode");
					break;
				case STOREMODETOVFO:
					rc = storemodetovfo(to, from);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case SETDUPLEX:
					rc = setduplex(to, from, cmdlopts[i].iarg);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case SETTXPL:
					rc = setpl(to, from, cmdlopts[i].iarg, 0);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case SETRXPL:
					rc = setpl(to, from, cmdlopts[i].iarg, 1);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case GETVFOCODEPLUG:
					rc = getvfocodeplug(to, from, inbuf, 255);
					if (rc) perror("xcatctl: storemodetovfo");
					else
					{
						if (verbose)
							fputs("VFO code plug: ", stdout);
						printf("%02x %02x %02x %02x %02x %02x %02x %02x "
							"%02x %02x %02x %02x %02x %02x %02x %02x\n",
							inbuf[0], inbuf[1], inbuf[2], inbuf[3],
							inbuf[4], inbuf[5], inbuf[6], inbuf[7],
							inbuf[8], inbuf[9], inbuf[10], inbuf[11],
							inbuf[12], inbuf[13], inbuf[14], inbuf[15]);
					}
					break;
				case SETVFOCODEPLUG:
					rc = setvfocodeplug(to, from, cmdlopts[i].sarg);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case GETVERSION:
					rc = getversion(to, from, inbuf, 255);
					if (rc) perror("xcatctl: storemodetovfo");
					else puts(inbuf);
					break;
				case GETCONFIG:
					rc = getconfig(to, from, inbuf, 255);
					if (rc) perror("xcatctl: storemodetovfo");
					else
					{
						if (verbose)
							decodeconfig(inbuf[0], inbuf[1]);
						else
							printf("%02x %02x\n", inbuf[0], inbuf[1]);
					}
					break;
				case SETCONFIG:
					rc = setconfig(to, from, confdata0, confdata1);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case SETCONFIGBITS:
					a = (cmdlopts[i].iarg & 0xff00) >> 8;
					b = (cmdlopts[i].iarg & 0xff);
					rc = setconfig(to, from, a, b);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case GETTXOFFSET:
					rc = gettxoffset(to, from, &offset);
					if (rc) perror("xcatctl: storemodetovfo");
					else
					{
						if (verbose)
							printf("Transmit offset is %7.3f MHz\n",
								(float)(offset / 1000000.0));
						else
							printf("%ld\n", offset);
					}
					break;
				case SETTXOFFSET:
					rc = settxoffset(to, from, cmdlopts[i].iarg);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case GETVCOSPLIT:
					rc = getvcosplitfreq(to, from, &txsplit, &rxsplit);
					if (rc) perror("xcatctl: storemodetovfo");
					else
					{
						if (verbose)
							printf("Transmit VCO split is %7.3f MHz; "
								"receive VCO split is %7.3f MHz\n",
								(double)(txsplit / 1000000.0),
								(double)(rxsplit / 1000000.0));
						else
							printf("%7.3f %7.3f\n",
								(double)(txsplit / 1000000.0),
								(double)(rxsplit / 1000000.0));
					}
					break;
				case SETVCOSPLIT:
					if ((! cmdlopts[TXVCOSPLIT].oflag) ||
						(! cmdlopts[RXVCOSPLIT].oflag))
					{
						fputs("xcatctl: You must specify --txvcosplit and "
							"--rxvcosplit with --setvcosplit\n", stderr);
						exit(EINVAL);
					}
					rc = setvcosplit(to, from, cmdlopts[TXVCOSPLIT].iarg,
							cmdlopts[RXVCOSPLIT].iarg);
					if (rc) perror("xcatctl: storemodetovfo");
					break;
				case TXVCOSPLIT:
					break;
				case RXVCOSPLIT:
					break;
				case RADIOBAND:
					break;
				case SENDCOS:
					break;
				case CONTROLSYS:
					break;
				case IO0:
					break;
				case IO1:
					break;
				case IO2:
					break;
				case IO3:
					break;
				case IO4:
					break;
				case IO5:
					break;
				case IO6:
					break;
				case IO7:
					break;
				case TOADDR:
					break;
				case FROMADDR:
					break;
				case DEBUG:
					break;
				case SERIALPORT:
					break;
				case HELP:
					break;
				case SHOWVERSION:
					break;
				default:
					fprintf(stderr, "xcatctl: logic error in dispatcher, "
						"index %d\n", i);
					exit(EINVAL);
			}
		}
	}

/* Done */

	serclose();
	exit(0);
}
